
import sys
import random
from . import SimUtils
from SimulatorFramework.SimCanvasRubberband import SimCanvasRubberband
from . import Frontier
from . import TileAssembly
from SimulatorFramework.TileType import TileType, Tile
from SimulatorFramework.TileTypeEditor import *
if sys.version_info[0] < 3:
    import thread
    from wx.lib.pubsub import pub
else:
    from pubsub import pub
import threading
import wx
import wx.adv
import time


CELLSIZE  = 60 #96

SIMFW_FULLDISPLAY      = 0
#SIMFW_GROWTHDISPLAY    = 1
SIMFW_NODISPLAY        = 1

SIMFW_MODE_PAN_ZOOM    = 0
SIMFW_MODE_SELECT      = 1
SIMFW_MODE_DRAW        = 2
SIMFW_MODE_TILE_ADDREM = 3

SIMFW_DRAWQUEUE_DRAWTILE        = 0
SIMFW_DRAWQUEUE_REMOVETILE      = 1
SIMFW_DRAWQUEUE_DRAWFRONTIER    = 2
SIMFW_DRAWQUEUE_REMOVEFRONTIER  = 3


class Canvas(wx.ScrolledCanvas):
    def __init__(self, parent, idIn, frame, size, app):
        wx.ScrolledCanvas.__init__(self, parent, idIn, (0, 0), size=size, style=wx.SUNKEN_BORDER)

        self.frame = frame
        self.app = app
        self.lines = []
        self.drawScale = 1.0
        self.DrawQueue = []
        self.needsRedraw = False
        self.RemoveQueue = []
        self.drawUpdates = SIMFW_FULLDISPLAY
        self.CreateNewAssembly()
        self.drawGrid = False
        self.keepGoing = False
        self.showMouseOver = True
        self.displayPatterns = []
        self.breakpoints = []
        self.currentBreakpoint = -1
        self._lock = threading.Lock()
        self.breakOnNondeterminism = False
        self.savedCenteredCoords = None
        self.dragStart = None
        self.grabbingCursorImg = wx.Image("images/grabbing.gif", wx.BITMAP_TYPE_GIF)
        self.grabCursorImg = wx.Image("images/grab.gif", wx.BITMAP_TYPE_GIF)
        self.backgroundColor = wx.Colour((255, 255, 255, 255))
        self.backgroundBrush = wx.Brush(self.backgroundColor)
        self.backgroundPen = wx.Pen(self.backgroundColor)
        self.imgOffsetHoriz = 0
        self.imgOffsetVert = 0
        
        #Can't do this!  It causes a full redraw (i.e. paint) after every draw
        #self.SetDoubleBuffered(True)

        self.SetRandomSeed()

        # create a PseudoDC to record our drawing
        #self.pdc = wx.PseudoDC()

        self.pen_cache = {}
        self.brush_cache = {}
        self.gridPen = self.ColoredPen("BLACK", 1)
        self.greenBrush = self.ColoredBrush("GREEN")
        self.redBrush = self.ColoredBrush("RED")
        self.whiteBrush = self.ColoredBrush("WHITE")
        self.yellowBrush = self.ColoredBrush("YELLOW")
        self.whitePen = self.ColoredPen("WHITE", 1)
        self.SetBackgroundColour(self.backgroundColor)

        self.SetVirtualSize(int(self.width*self.drawScale), int(self.height*self.drawScale))

        self.lastRect = wx.Rect(-999999999, -999999999, 0, 0)
        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_ERASE_BACKGROUND, lambda x:None)
        self.Bind(wx.EVT_MOUSE_EVENTS, self.OnMouse)
        self.Bind(wx.EVT_CHAR,self.app.OnChar)
        
        self.SubscribeEventHandlers()

        self.mouseMode = SIMFW_MODE_PAN_ZOOM
        self.SetCursor(wx.Cursor(self.grabCursorImg))

        self.InitializeDrawing()
        self.SetScrollRate(1,1)
        self.ScrollToCoords((0,0))
        
        self.rubberBand = SimCanvasRubberband.SimCanvasRubberband(drawingSurface=self)
        self.rubberBand.reset(aspectRatio=0.5)
        self.rubberBand.reset()
        self.lastDeletedList = []

    def SubscribeEventHandlers(self):
        pub.getDefaultPublisher().subscribe(self.UpdateStatusbar, "updateStatusbar")
        pub.getDefaultPublisher().subscribe(self.DoDrawAssemblyUpdate, "drawAssembly")
        pub.getDefaultPublisher().subscribe(self.DoCanvasSizeUpdate, "sizeCanvas")
        
    def UnsubscribeEventHandlers(self):
        pub.getDefaultPublisher().unsubscribe(self.UpdateStatusbar, "updateStatusbar")
        pub.getDefaultPublisher().unsubscribe(self.DoDrawAssemblyUpdate, "drawAssembly")
        pub.getDefaultPublisher().unsubscribe(self.DoCanvasSizeUpdate, "sizeCanvas")
        
    def UpdateSizeVars(self):
        self.width  = int(self.assembly.GetWidth() * self.sideLen)
        self.height = int(self.assembly.GetHeight() * self.sideLen)
        self.numVert = int(self.width / self.sideLen)
        self.numHoriz = int(self.height / self.sideLen)

    def CreateNewAssembly(self):
        self.sideLen = CELLSIZE
        self.assembly = TileAssembly.Assembly(tileWidths=self.sideLen, tileHeights=self.sideLen)
        self.UpdateSizeVars()
        #self.RedrawAssembly()
        self.CanvasSizeUpdate((0,0))

    def SendStatusBarUpdate(self, field, msg):
        wx.CallAfter(pub.getDefaultPublisher().sendMessage, "updateStatusbar", msg = (field, msg))

    def UpdateStatusbar(self, msg):
        self.frame.SetStatusText(msg[1], msg[0])

    def SetRandomSeed(self, seed=-1):
        self.randSeed = seed
        if self.randSeed < 0:
            self.randSeed = random.randint(0, sys.maxsize)
        SimUtils.globalApp.OutputMessage("Seeding random number generator with " + str(self.randSeed))
        random.seed(self.randSeed)

    def DoSetAsSeed(self):
        self.assembly.SetCurrentAssemblyAsSeed()

    def DoLoad(self):
        openFileDialog = wx.FileDialog(self.app.window, "Open", "", "", 
                                       "Tile system files (*.txt)|*.txt",
                                       wx.FD_OPEN | wx.FD_FILE_MUST_EXIST)
        openFileDialog.ShowModal()
        fileName = openFileDialog.GetPath()
        openFileDialog.Destroy()
        if (len(fileName) > 0):
            
            dlg = wx.TextEntryDialog(self, "You can set your random seed below\n(use -1 to choose a new random value):", caption="Set random seed", value=str(self.randSeed))
            if dlg.ShowModal() == wx.ID_OK:
                self.randSeed = int(dlg.GetValue())
                dlg.Destroy()
            else:
                dlg.Destroy()
                return

            self.SetRandomSeed(self.randSeed)
            SimUtils.globalApp.OutputMessage("Loading tile assembly system...")
            self.InitializeDrawing()
            self.CreateNewAssembly()
            midPt = self.assembly.LoadTileSystem(fileName)
            if midPt == False:
                self.app.OutputMessage("Failed to load tile assembly system from " + fileName)
                return
            self.RedrawAssembly(checkAll=True)
            self.ScrollToCoords(midPt)
            SimUtils.globalApp.OutputMessage("Tile assembly system loaded")
            
    def DoLoadAssembly(self):
        SimUtils.globalApp.OutputMessage("Sorry, this isn't implemented")

    def DoSave(self):
        #saveFileDialog = wx.FileDialog(self.window, "Save As", "", "", 
        #                               "Tile system files (*.txt)|*.txt",
        #                               wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
        #saveFileDialog.ShowModal()
        #fileName = saveFileDialog.GetPath()
        #saveFileDialog.Destroy()

        #self.saveSnapshot()

        dlg = wx.MessageDialog(self, "Sorry, this hasn't been implemented yet..")
        dlg.ShowModal()
        dlg.Destroy()
        return
    
        currFile = "output.txt"
        newFile = wx.FileSelector("Save assembly to file...", currFile, currFile, ".txt", "*.txt", wx.FD_SAVE|wx.FD_OVERWRITE_PROMPT)
        if len(newFile) > 0:
            fOut = open(newFile, "w")
            iCount = 0
            jCount = 0
            for i in range(self.numHoriz):
                for j in range(self.numVert):
                    if self.grid.IsEmpty(i, j) == False:
                        pt = self.grid.GetCenterOffsetOfType(i, j)
                        if pt.x == 0 and pt.y == 0:
                            name = self.grid.GetType(i, j)
                            outStr = name + "," + str(i) + "," + str(j) + "\n"
                            fOut.write(outStr)
            fOut.close()
            
    def DoSaveAs(self):
        dlg = wx.MessageDialog(self, "Sorry, this hasn't been implemented yet..")
        dlg.ShowModal()
        dlg.Destroy()
        return

    def saveSnapshot(self, dcSource):
        
        dlg = wx.MessageDialog(self, "Sorry, this hasn't been implemented yet..")
        dlg.ShowModal()
        dlg.Destroy()
        return

        # based largely on code posted to wxpython-users by Andrea Gavana 2006-11-08
        #size = dcSource.Size
    
        # Create a Bitmap that will later on hold the screenshot image
        # Note that the Bitmap must have a size big enough to hold the screenshot
        # -1 means using the current default colour depth
#         bmp = wx.EmptyBitmap(self.width, self.height)
#     
#         # Create a memory DC that will be used for actually taking the screenshot
#         memDC = wx.MemoryDC()
#     
#         # Tell the memory DC to use our Bitmap
#         # all drawing action on the memory DC will go to the Bitmap now
#         memDC.SelectObject(bmp)
#     
#         # Blit (in this case copy) the actual screen on the memory DC
#         # and thus the Bitmap
#         memDC.Blit( 0, # Copy to this X coordinate
#             0, # Copy to this Y coordinate
#             self.width, # Copy this width
#             self.height, # Copy this height
#             dcSource, # From where do we copy?
#             0, # What's the X offset in the original DC?
#             0  # What's the Y offset in the original DC?
#             )
        
        bmp = wx.EmptyBitmap(self.width, self.height)
        #wdc = wx.WindowDC(self)
        memDC = wx.MemoryDC(bmp)

        bg = wx.Brush("WHITE")
        memDC.SetBackground(bg)
        memDC.Clear()
        #memDC.Blit(0, 0, self.width, self.height, wdc, 0, 0)
        
        for i in range(self.assembly.minX, self.assembly.maxX+1):
            for j in range(self.assembly.minY, self.assembly.maxY+1):
                loc = self.assembly.GetLoc((i,j))
                if loc is not None:
                    if self.NameMatchesPatterns(loc.type.name) == True:
                        self.DrawAtCoords(memDC, i, j, self.assembly.GetLoc((i,j)))
        
        # Select the Bitmap out of the memory DC by selecting a new
        # uninitialized Bitmap
        memDC.SelectObject(wx.NullBitmap)
        imageFileName = 'snapshots-saved.bmp'
        bmp.SaveFile(imageFileName, wx.BITMAP_TYPE_BMP)
        
        ##pdfFileName = 'snapshots-saved.pdf'
        #=======================================================================
        # c = canvas.Canvas(pdfFileName, pagesize=letter)
        # width, height = letter
        # c.drawImage(imageFileName, inch, height - 2 * inch) # Who needs consistency?
        # c.showPage()
        # c.save()
        #=======================================================================
        #=======================================================================
        # doc = SimpleDocTemplate(pdfFileName, pagesize=(self.width,self.height))
        # parts = []
        # parts.append(Image(imageFileName))
        # doc.build(parts)
        #=======================================================================
    
        #img = bmp.ConvertToImage()
        #img.SaveFile('snapshots-saved.png', wx.BITMAP_TYPE_PNG)

    def Reset(self, bFromFiles=True):
        dlg = wx.TextEntryDialog(self, "Are you sure you want to reset your system\nand abandon the current assembly?\nIf so, you can set your random seed below\n(use -1 to choose a new random value):", caption="Set random seed", value=str(self.randSeed))
        if not dlg.ShowModal() == wx.ID_OK:
            return
        result = int(dlg.GetValue())
        dlg.Destroy()
        self.SetRandomSeed(result)

        self.app.OutputMessage("Reseting system...")

        midPt = (0,0)
        if bFromFiles == True:
            if not self.systemDefFile is None:
                midPt = self.assembly.LoadTileSystem(self.systemDefFile)
        else:
            self.assembly.ClearSpace()
            midPt = self.assembly.PlaceSeed()
        self.UpdateSizeVars()
        self.RedrawAssembly(checkAll=True)
        self.CanvasSizeUpdate(midPt)
        self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.numTiles) + ", frontier size: " + str(len(self.assembly.frontier)))
        
        self.app.OutputMessage("Tile assembly system reset")
        
    def DoRewindStep(self):
        self.app.OutputMessage("Sorry... not yet implemented!")
        return False
    
        retVal = False
        if self.assembly.bKeepCache == True:
            rewindStep = self.assembly.RewindCache()
            if rewindStep[0] == True:
                self.AddToDrawQueue(rewindStep[1])
                retVal = True
        if self.drawUpdates == SIMFW_NODISPLAY:
            if (self.assembly.numTiles % 1000) == 0:
                #self.app.OutputMessage("Number of tiles in assembly: " + str(self.assembly.numTiles))
                self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.numTiles) + ", frontier size: " + str(len(self.assembly.frontier)))
        elif (self.keepGoing == False) or (self.assembly.numTiles % 1000) == 0:
            #MJP TODO: handle this better, but it REALLY slows things down!!!
            self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.numTiles) + ", frontier size: " + str(len(self.assembly.frontier)))
        return retVal

    def GoBackward(self):
        self.showMouseOver = False
        self.SendStatusBarUpdate(1, "")
        while self.keepGoing == True:
            if self.DoRewindStep() == False:
                self.keepGoing = False
        SimUtils.PrintDebug("Stopping the thread!", SimUtils.MSG_MSG)
        self.showMouseOver = True
        self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.numTiles) + ", frontier size: " + str(len(self.assembly.frontier)))

    def DoGoBackward(self):
        if self.keepGoing == True:
            self.app.OutputMessage("ERROR: Cannot create another simulation thread, there is already one running!")
            return
        self.keepGoing = True
        try:
            self.goThread = threading.Thread(target=self.GoBackward)
            self.goThread.start()
            # self.goThread = thread.start_new_thread(self.GoBackward, ())
        except:
            self.app.OutputMessage("ERROR: Failed to create thread!")

    def DoGo(self):
        if self.keepGoing == True:
            self.app.OutputMessage("ERROR: Cannot create another simulation thread, there is already one running!")
            return
        self.keepGoing = True
        try:
            self.goThread = threading.Thread(target=self.GoForward)
            self.goThread.start()
            # self.goThread = thread.start_new_thread(self.GoForward, ())
        except:
            self.app.OutputMessage("ERROR: Failed to create thread!")
    
    def DoQuit(self):
        self.keepGoing = False
        self.showMouseOver = True
        self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.GetNumTiles()) + ", frontier size: " + str(len(self.assembly.frontier)))
    
    def GoForward(self):
        self.showMouseOver = False
        self.SendStatusBarUpdate(1, "")
        while self.keepGoing == True:
            if self.DoForwardStep() == False:
                self.keepGoing = False
        SimUtils.PrintDebug("Stopping the thread!", SimUtils.MSG_MSG)
        self.showMouseOver = True
        self.SendStatusBarUpdate(1, "Number of polyominoes: " + str(self.assembly.GetNumTiles()))

    def DoForwardStep(self):
        savedMinX = self.assembly.minX
        savedMinY = self.assembly.minY
        self._lock.acquire()
        assemblyUpdates = self.assembly.StepForward()
        self._lock.release()
        if not assemblyUpdates is None:
            if not self.width == (self.assembly.width * self.sideLen) or not self.height == (self.assembly.height * self.sideLen):
                centerCoords = self.GetCenterCoords()
                coords = (centerCoords[0] + savedMinX - self.assembly.minX, centerCoords[1] + savedMinY - + self.assembly.minY)
                #TODO: There is a bug related to recentering after canvas growth. Perhaps it is related to additional growth of the assembly before canvas resizing
                print(f'{time.time()} About to update canvas size, centerCoords = {centerCoords}, savedMins = {(savedMinX,savedMinY)}, assembly mins = {(self.assembly.minX,self.assembly.minY)}, computed coords = {coords}')
                self.UpdateSizeVars()
                self.CanvasSizeUpdate(coords)
            else:
                self.AddToDrawQueue(assemblyUpdates[0])
            self.CheckForBreakpointHit(assemblyUpdates[0][0][0])
            if self.drawUpdates == SIMFW_NODISPLAY:
                if (self.assembly.numTiles % 1000) == 0:
                    
                    self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.numTiles) + ", frontier size: " + str(len(self.assembly.frontier)))
            elif (self.keepGoing == False) or (self.assembly.numTiles % 1000) == 0:
                #MJP TODO: handle this better, but it REALLY slows things down!!!
                self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.numTiles) + ", frontier size: " + str(len(self.assembly.frontier)))
            return assemblyUpdates[1]
        return False
    
    def CheckForBreakpointHit(self, newTile):
        if len(self.breakpoints) > 0:
            for bp in self.breakpoints:
                if bp.type == SimUtils.SIMU_BP_TYPE_TILENUM:
                    if self.assembly.numTiles == bp.tileNum:
                        self.keepGoing = False
                        SimUtils.globalApp.OutputMessage("Hit breakpoint for " + str(self.assembly.numTiles) + " tiles in the assembly.")
                elif bp.type == SimUtils.SIMU_BP_TYPE_TILETYPE:
                    if bp.tileType == newTile.type.name:
                        self.keepGoing = False
                        SimUtils.globalApp.OutputMessage("Hit breakpoint for tile of type " + bp.tileType + " placed in the assembly.")
                else:
                    newCoords = newTile.coords
                    bHit = False
                    bKeepChecking = True
                    if (bp.tileLoc[0] is not None):
                        if (bp.tileLoc[0] == newCoords[0]):
                            bHit = True
                        else:
                            bKeepChecking = False
                    if (bKeepChecking == True) and (bp.tileLoc[1] is not None):
                        if (bp.tileLoc[1] == newCoords[1]):
                            bHit = True
                        else:
                            bHit = False
                            bKeepChecking = False
                    if (bKeepChecking == True) and (bp.tileLoc[2] is not None):
                        if (bp.tileLoc[2] == newCoords[2]):
                            bHit = True
                        else:
                            bHit = False
                    if bHit == True:
                        self.keepGoing = False
                        SimUtils.globalApp.OutputMessage("Hit breakpoint for tile placed in location " + str(newCoords) + " which matches " + str(bp.tileLoc))

    def GetCenterCoords(self):
        usp = self.CalcUnscrolledPosition(0,0)
        sz = self.GetSize()
        center = (usp[0] + int(sz[0]/2), usp[1] + int(sz[1]/2))
        coords = self.CoordsFromPixelPos(center[0], center[1], True)
        return coords
    
    def ConvertEventCoords(self, event):
        xView, yView = self.GetViewStart()
        xDelta, yDelta = self.GetScrollPixelsPerUnit()
        return (event.GetX() + (xView * xDelta),
            event.GetY() + (yView * yDelta))

    def OffsetRect(self, r):
        xView, yView = self.GetViewStart()
        xDelta, yDelta = self.GetScrollPixelsPerUnit()
        r.OffsetXY(-(xView*xDelta),-(yView*yDelta))

    def GetSelectedCells(self):
        if (self.rubberBand is None) or (self.rubberBand.currentBox is None):
            return (0,0,0,0)

        xView, yView = self.GetViewStart()
        xDelta, yDelta = self.GetScrollPixelsPerUnit()

        (x1,y1,x2,y2) = self.rubberBand.currentBox
        (xLeft, yTop) = (x1 + (xView * xDelta), y1 + (yView * yDelta))
        (xRight, yBottom) = (x1 + x2 + (xView * xDelta), y1 + y2 + (yView * yDelta))

        cLeft, cTop = self.CoordsFromPixelPos(xLeft, yTop)
        cRight, cBottom = self.CoordsFromPixelPos(xRight, yBottom)
        
        return (cLeft, cTop, cRight, cBottom)
        
    def DeleteSelection(self, bInside=True):
        (cLeft, cTop, cRight, cBottom) = self.GetSelectedCells()
        self.lastDeletedList = []
        if bInside == True:
            for x in range(cLeft+1, cRight):
                for y in range(cBottom+1, cTop):
                    loc = self.assembly.GetLoc((x,y))
                    if not loc is None:
                        if self.assembly.RemoveTile((x,y)) == True:
                            self.lastDeletedList.append(loc)
        else:
            for x in range(self.assembly.minX, self.assembly.maxX+1):
                for y in range(self.assembly.minY, self.assembly.maxY+1):
                    if (x <= cLeft) or (x >= cRight) or (y <= cBottom) or (y >= cTop):
                        loc = self.assembly.GetLoc((x,y))
                        if not loc is None:
                            if self.assembly.RemoveTile((x,y)) == True:
                                self.lastDeletedList.append(loc)
        self.rubberBand.reset()
        self.DoRedrawAssembly()
        
    def CopySelection(self):
        (cLeft, cTop, cRight, cBottom) = self.GetSelectedCells()
        copiedTiles = []
        for y in reversed(range(cBottom+1, cTop)):
            for x in range(cLeft+1, cRight):
                loc = self.assembly.GetLoc((x,y))
                if not loc is None:
                    adjCoords = (loc.coords[0]-cLeft, loc.coords[1]-cTop)
                    copiedTiles.append(Tile(loc.type, adjCoords))
        assemblyString = self.assembly.AssemblyToString(copiedTiles)
        if not wx.TheClipboard.IsOpened():
            if wx.TheClipboard.Open():
                wx.TheClipboard.SetData(wx.TextDataObject(assemblyString))
                wx.TheClipboard.Close()
        
    def PasteSelection(self):
        assemblyString = ""
        if not wx.TheClipboard.IsOpened():
            if wx.TheClipboard.Open():
                data = wx.TextDataObject()
                wx.TheClipboard.GetData(data)
                assemblyString = data.GetText()
                wx.TheClipboard.Close()
        #MJP TODO: handle errors and report accordingly here!
        copiedTiles = self.assembly.AssemblyFromString(assemblyString)
        (cLeft, cTop, cRight, cBottom) = self.GetSelectedCells()
        for loc in copiedTiles:
            coords = (cLeft + int(loc[1][0]), cTop + int(loc[1][1]))
            self.assembly.AddTile(self.assembly.tileTypes[loc[0]], coords)
        self.rubberBand.reset()
        coords = self.GetCenterCoords()
        self.UpdateSizeVars()
        self.CanvasSizeUpdate(coords)

    def DisableSelectedFrontier(self):
        (cLeft, cTop, cRight, cBottom) = self.GetSelectedCells()
        for x in range(cLeft+1, cRight):
            for y in range(cBottom+1, cTop):
                fLoc = Frontier.FrontierLocation((x,y), None)
                #self.frontier.discard(fLoc)
                #self.frontier.add(fLoc)
                if fLoc in self.assembly.frontier:
                    self.assembly.disabledFrontier.add(fLoc)
                    self.assembly.SetLoc(fLoc.coords, TileAssembly.FRONTIER_LOC_OFF)
        self.rubberBand.reset()
        self.DoRedrawAssembly()
        
    def EnableSelectedFrontier(self):
        (cLeft, cTop, cRight, cBottom) = self.GetSelectedCells()
        for x in range(cLeft+1, cRight):
            for y in range(cBottom+1, cTop):
                fLoc = Frontier.FrontierLocation((x, y),"")
                if fLoc in self.assembly.disabledFrontier:
                    self.assembly.disabledFrontier.remove(fLoc)
                    if fLoc in self.assembly.frontier:
                        self.assembly.SetLoc(fLoc.coords, TileAssembly.FRONTIER_LOC)
        self.rubberBand.reset()
        self.DoRedrawAssembly()
        
    def MessageToggle(self, bShow):
        if bShow == True:
            self.app.showMessages = bShow
            self.app.OutputMessage("Setting display mode to display all messages.")
        else:
            self.app.OutputMessage("Setting display mode to hide all message updates.")
            self.app.showMessages = bShow

    def PrintMessage(self):
        print("This is a placeholder for printing out debugging or other useful info")

    def ZoomThenRecenter(self, zoomIn=True):
            self.DoScale(zoomIn, bScrollToCenter=True)

    def ToggleGrid(self):
            self.drawGrid = not self.drawGrid
            self.RedrawAssembly()
            self.Refresh()
    # Options are:
    #    SIMFW_FULLDISPLAY
    #    SIMFW_GROWTHDISPLAY
    #    SIMFW_NODISPLAY

    def ToggleDisplay(self):
            self.drawUpdates = (self.drawUpdates + 1) % 2
            if self.drawUpdates == SIMFW_FULLDISPLAY:
                if len(self.displayPatterns) == 0:
                    SimUtils.globalApp.OutputMessage("Switched display mode:  Displaying all tiles in the assembly")
                else:
                    SimUtils.globalApp.OutputMessage("Switched display mode:  Displaying all tiles matching the following patterns: " + str(self.displayPatterns))
                self.UpdateSizeVars()
                #self.RedrawAssembly()
                if not self.savedCenteredCoords is None:
                    self.CanvasSizeUpdate(self.savedCenteredCoords)
                self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.numTiles) + ", frontier size: " + str(len(self.assembly.frontier)))
            #elif self.drawUpdates == SIMFW_GROWTHDISPLAY:
            #    SimUtils.globalApp.OutputMessage("Switched display mode:  Displaying only addition of growth tiles")
            elif self.drawUpdates == SIMFW_NODISPLAY:
                self.savedCenteredCoords = self.GetCenterCoords()
                SimUtils.globalApp.OutputMessage("Switched display mode:  No tile additions are being displayed")

    def SetShowMatching(self):
        patternString = ""
        bDiff = False
        for pattern in self.displayPatterns:
            if len(patternString) > 0:
                patternString = patternString + ","
            patternString = patternString + pattern
        dlg = wx.TextEntryDialog(self, "Enter a comma separated list of strings.  Only tiles whose type\nnames contain at least one of these strings will be displayed.\nTo display all tile types, leave the text box blank.", caption="Set display patterns", value=str(patternString))
        if dlg.ShowModal() == wx.ID_OK:
            patternArray = []
            newPatterns = dlg.GetValue()
            newPatterns = str(newPatterns.strip())
            if len(newPatterns) > 0:
                patternArray = newPatterns.split(",")
            bDiff = (self.displayPatterns != patternArray)
            self.displayPatterns = patternArray
        dlg.Destroy()
        if bDiff == True:
            self.RedrawAssembly(checkAll=True)
            self.Refresh()
            
    def ShowHelp(self):
        info = wx.adv.AboutDialogInfo()
        info.SetName("Generic framework for a self-assembly simulator")
        info.SetVersion("0.1.0")
        helpStr = "\nList of commands:\n\n"
        helpStr = helpStr + "  l : select tile system definition file to load\n"
        helpStr = helpStr + "  s : save current assembly\n"
        helpStr = helpStr + "  r : reset current tile system to seed\n"
        helpStr = helpStr + "  f : simulate one step\n"
        helpStr = helpStr + "  g : begin fast forward simulation\n"
        helpStr = helpStr + "  q : quit fast forward simulation\n"
        helpStr = helpStr + "  p : print list of frontier locations\n"
        helpStr = helpStr + "  t : toggle display of grid lines\n"
        helpStr = helpStr + "  d : toggle display updates of tiles to assembly\n"
        helpStr = helpStr + " +/-: zoom in/out\n"
        helpStr = helpStr + "<left mouse click>:  zoom in on clicked location\n"
        helpStr = helpStr + "<right mouse click>:  zoom out of clicked location\n"
        helpStr = helpStr + "  h : show this help dialog"
        info.SetDescription(helpStr)
        wx.adv.AboutBox(info)

    def SetMouseOverText(self, coords):
        if coords == None:
            self.SendStatusBarUpdate(0, "")
            self.SendStatusBarUpdate(2, "")
            return
        tileTypeName = ""
        tile = self.assembly.GetLoc((coords[0], coords[1]))
        if not tile is None:
            tileTypeName = "Tile type: " + tile.type.ToString(bFull=True)
        else:
            floc = Frontier.FrontierLocation((coords[0], coords[1]),"")
            if floc in self.assembly.frontier:
                tileTypeName = "Frontier location"
                for f in self.assembly.frontier:
                    if f == floc:
                        tileTypeName = "Frontier location, allows " + str(f.fittingTypeNames)
            
        msg = "Coords: (" + str(int(coords[0])) + "," + str(int(coords[1])) + ")"
        self.SendStatusBarUpdate(0, msg)
        self.SendStatusBarUpdate(2, tileTypeName)

    def OnMouse(self, event):
        x,y = self.ConvertEventCoords(event)
        x2,y2 = self.CoordsFromPixelPos(x, y)

        if self.showMouseOver == True:
            if (x2 >= self.assembly.minX) and (x2 <= self.assembly.maxX) and (y2 >= self.assembly.minY) and (y2 <= self.assembly.maxY):
                self.SetMouseOverText((x2, y2))
            else:
                self.SetMouseOverText(None)
            
        if self.mouseMode == SIMFW_MODE_SELECT:
            if event.RightUp():
                self.rubberBand.reset()
            else:
                self.rubberBand.HandleMouseEvents(event)
        elif self.mouseMode == SIMFW_MODE_TILE_ADDREM:
            if event.RightDown():
  
                tile = self.assembly.GetLoc((x2, y2))
                if tile is None:
                    popupSize = wx.Size(300,300)
                    popup = TileTypePopup(self, self.assembly.tileTypes, self.frame, self.app, popupSize, (x2,y2))
                else:
                    popupSize = wx.Size(200,50)
                    popup = DeleteTilePopup(self, None, self.frame, self.app, popupSize, (x2,y2))
                (screenX, screenY) = self.ClientToScreen(event.GetX(), event.GetY())
                popupPos = (screenX-popupSize[0], screenY-popupSize[1])
                popup.Position(ptOrigin=popupPos, sizePopup=popupSize)
                popup.Popup()
                
        elif self.mouseMode == SIMFW_MODE_DRAW:
            if event.LeftDown() or (event.LeftIsDown() and event.Dragging()):
                loc = self.assembly.GetLoc((x2,y2))
                if loc == None:
                    fLoc = Frontier.FrontierLocation((x2,y2), None)
                    #if fLoc in self.assembly.frontier:
                    self.assembly.disabledFrontier.add(fLoc)
                    self.assembly.SetLoc(fLoc.coords, TileAssembly.FRONTIER_LOC_OFF)
                    self.AddToDrawQueue(([], [], [fLoc], []))
            elif event.RightDown() or (event.RightIsDown() and event.Dragging()):
                fLoc = Frontier.FrontierLocation((x2,y2), None)
                if fLoc in self.assembly.disabledFrontier:
                    self.assembly.disabledFrontier.discard(fLoc)
                    if fLoc in self.assembly.frontier:
                        self.assembly.SetLoc(fLoc.coords, TileAssembly.FRONTIER_LOC)
                        self.AddToDrawQueue(([], [], [fLoc], []))
                    else:
                        self.assembly.ClearLoc(fLoc.coords)
                        self.AddToDrawQueue(([], [], [], [fLoc]))
        else:
            if event.LeftDown():
                if self.mouseMode == SIMFW_MODE_PAN_ZOOM:
                    self.SetCursor(wx.Cursor(self.grabbingCursorImg))
                    self.dragStart = self.ScreenToClient(wx.GetMousePosition())
                    self.dragStartScrollPos = (self.GetScrollPos(wx.HORIZONTAL), self.GetScrollPos(wx.VERTICAL))
    
            elif event.RightDown():
                if self.mouseMode == SIMFW_MODE_PAN_ZOOM:
                    
                    bZoomIn = True
                    if (event.AltDown() == True) or (event.ControlDown() == True) or (event.ShiftDown() == True):
                        bZoomIn = False
                    self.DoScale(bZoomIn, bScrollToCenter=False, scrollToCoords=(x2,y2))
                    #self.ScrollToCoords((x2,y2))
                    sz = self.GetSize()
                    self.WarpPointer(int(sz[0]/2), int(sz[1]/2))

            elif event.LeftUp():
                self.dragStart = None
                if self.mouseMode == SIMFW_MODE_PAN_ZOOM:
                    self.SetCursor(wx.Cursor(self.grabCursorImg))
                else:
                    self.SetCursor(wx.Cursor(wx.CURSOR_ARROW))
            
            elif event.Dragging():
                if self.mouseMode == SIMFW_MODE_PAN_ZOOM:
                    if self.dragStart is not None:
                        #self.SetScrollRate(1,1)
                        new = self.ScreenToClient(wx.GetMousePosition())
                        dx = int((self.dragStart.x-new.x)/1)
                        dy = int((self.dragStart.y-new.y)/1)
                        spx = self.GetScrollPos(wx.HORIZONTAL)
                        spy = self.GetScrollPos(wx.VERTICAL)
                        self.Scroll(spx+dx, spy+dy)
                        self.dragStart = new


    def ColoredPen(self, color, width = 2):
        if not (color, width) in self.pen_cache:
            self.pen_cache[(color, width)] = wx.Pen(color, width)
        return self.pen_cache[(color, width)]

    def ColoredBrush(self, color):
        if not color in self.brush_cache:
            self.brush_cache[color] = wx.Brush(color)
        return self.brush_cache[color]

    def SetBackgroundColor(self):
        if self.app.useAdvancedColorPicker == False:
            dlg = wx.ColourDialog(self)
     
            # Ensure the full colour dialog is displayed, not the abbreviated version.
            dlg.GetColourData().SetChooseFull(True)
     
            if dlg.ShowModal() == wx.ID_OK:
                data = dlg.GetColourData()
                self.backgroundColor = wx.Colour(data.GetColour().Get())
                #self.app.canvas.InitializeDrawing(self.app.canvas.pdc)
                self.assembly.CreateTileTypeImages(self.backgroundColor)
                self.Refresh()
            dlg.Destroy()
        else:
            cd = wx.ColourData()
            if type(self.backgroundColor) is str:
                cd.SetColour(wx.TheColourDatabase.Find(self.backgroundColor))
            else:
                cd.SetColour(self.backgroundColor)
            dlg = CCD.CubeColourDialog(self, cd)
            if dlg.ShowModal() == wx.ID_OK:
                colourData = dlg.GetColourData()
                colorTuple = (colourData.GetColour().Red(), colourData.GetColour().Green(), colourData.GetColour().Blue(), colourData.GetColour().Alpha())
                self.backgroundColor = wx.Colour(colorTuple)
                #box.beadType.color = str(colourData.GetColour().Get())
                #self.app.canvas.InitializeDrawing(self.app.canvas.pdc)
                self.assembly.CreateTileTypeImages(self.backgroundColor)
                self.Refresh()
            dlg.Destroy()
        self.backgroundBrush = wx.Brush(self.backgroundColor)
        self.backgroundPen = wx.Pen(self.backgroundColor)

    def OnPaint(self, event):
        self._lock.acquire()

        dc = wx.PaintDC(self)

        bg = wx.Brush(self.backgroundColor)
        dc.SetBackground(bg)
        dc.Clear()

        # use PrepareDC to set position correctly
        self.PrepareDC(dc)
        # we need to clear the dc BEFORE calling PrepareDC
        
        clipped = dc.GetClippingBox()

        # On OS X, PrepareDC doesn't seem to account for scrolling as it does on Windows
        #if "wxOSX" in wx.PlatformInfo:
        if False: # wxpython was recently updated so that the OS specific code is not needed
                  # for Windows and Linux, and I'm guessing maybe not for OS X either
            xView,yView = self.GetViewStart()
            xDelta,yDelta = self.GetScrollPixelsPerUnit()
            x = clipped[0] + xView * xDelta
            y = clipped[1] + yView * yDelta
        else:
            x, y = self.GetViewStart()
            w, h = self.GetSize()
            topLeft = self.CoordsFromPixelPos(x, y)
            bottomRight = self.CoordsFromPixelPos(x + w, y + h)
        
        dc.SetUserScale(self.drawScale, self.drawScale)
        
        if self.drawGrid == True:
            self.DrawGrid(dc)
        
        #for i in range(self.assembly.minX, self.assembly.maxX+1):
        #    for j in range(self.assembly.minY, self.assembly.maxY+1):
        for i in range(topLeft[0], bottomRight[0]+1):
            for j in range(bottomRight[1], topLeft[1]+1):
                loc = self.assembly.GetLoc((i,j), includeFrontier=True)
                if loc is not None:
                    if loc == TileAssembly.FRONTIER_LOC or loc == TileAssembly.FRONTIER_LOC_OFF:
                        self.DrawFrontierLocation(dc, (i,j), loc)
                    elif self.NameMatchesPatterns(loc.type.name) == True:
                        self.DrawAtCoords(dc, i, j, loc)
                
        self._lock.release()
        self.needsRedraw = False
        self.SendStatusBarUpdate(1, "Number of tiles: " + str(self.assembly.GetNumTiles()) + ", frontier size: " + str(len(self.assembly.frontier)))
        return

    def DrawGrid(self, dc):
        dc.SetPen(self.gridPen)
        for i in range(0, self.numVert+1):
            x = i * self.sideLen
            dc.DrawLine(x, self.height, x, self.height-self.numVert*self.sideLen)
        for i in range(0, self.numHoriz+1):
            y = i * self.sideLen
            dc.DrawLine(0, self.height-y, self.numHoriz*self.sideLen, self.height-y)

    def ScrollToCoords(self, coords):
        (x,y) = self.GetPixelsForCoords(coords[0], coords[1])
        x = x - self.sideLen/2
        y = y - self.sideLen/2
        x = x * self.drawScale
        y = y * self.drawScale
        (xs, ys) = self.GetScrollPixelsPerUnit()
        (xsz,ysz) = self.GetSize()
        self.Scroll(int((x-xsz/2+self.sideLen*self.drawScale)/xs), int((y-ysz/2+self.sideLen*self.drawScale)/ys))

    def AddToDrawQueue(self, fourTuple):
        (addedTiles, removedTiles, addedFrontierLocs, removedFrontierLocs) = fourTuple
        self._lock.acquire()
        for rt in removedTiles:
            self.DrawQueue.append((rt.coords, SIMFW_DRAWQUEUE_REMOVETILE))
        for rf in removedFrontierLocs:
            self.DrawQueue.append((rf.coords, SIMFW_DRAWQUEUE_REMOVEFRONTIER))
        for at in addedTiles:
            if self.NameMatchesPatterns(at.type.name) == True:
                self.DrawQueue.append((at.coords, SIMFW_DRAWQUEUE_DRAWTILE))
        for af in addedFrontierLocs:
            self.DrawQueue.append((af.coords, SIMFW_DRAWQUEUE_DRAWFRONTIER))
        self._lock.release()
        self.DrawAssemblyUpdate("new")
        
    def CanvasSizeUpdate(self, centerCoords=None):
        #if centerCoords == None:
        #    centerCoords = self.GetCenterCoords()
        if self.drawUpdates == SIMFW_NODISPLAY:
            return
        wx.CallAfter(pub.getDefaultPublisher().sendMessage, "sizeCanvas", coords = centerCoords)
        
    def DoCanvasSizeUpdate(self, coords=None):
        if coords is None:
            coords = self.GetCenterCoords()
        self.SetVirtualSize(int(self.width*self.drawScale), int(self.height*self.drawScale))
        self.ScrollToCoords(coords)
        self.Refresh()

    def DrawAssemblyUpdate(self, msg):
        if self.drawUpdates == SIMFW_NODISPLAY:
            return
        if self.needsRedraw == False:
            self.needsRedraw = True
            wx.CallAfter(pub.getDefaultPublisher().sendMessage, "drawAssembly", msg = msg)
        
    def DoDrawAssemblyUpdate(self, msg):
        if self.drawUpdates == SIMFW_NODISPLAY:
            return

        if msg == "all":
            self.Refresh()
            return
        elif msg == "allAndCheck":
        #    self.DoInitializeDrawing(dc)
            self.DoRedrawAssembly()

        self._lock.acquire()
               
         #HACK to work around a bug writing to the DC multiple times on OSX
        if not "wxOSX" in wx.PlatformInfo:
            dc = wx.ClientDC(self)
            self.PrepareDC(dc)
            # we need to clear the dc BEFORE calling PrepareDC
            dc.SetUserScale(self.drawScale, self.drawScale)

        #print("in drawupdate, queue len: " + str(len(self.DrawQueue)))
        if (msg == "new") and self.drawBitmaps == True:

            usp = self.CalcUnscrolledPosition(0,0)
            sz = self.GetSize()
            topLeft = self.CoordsFromPixelPos(usp[0], usp[1], True)
            bottomRight = self.CoordsFromPixelPos(usp[0]+sz[0], usp[1]+sz[1], True)

            for e in self.DrawQueue:
                
                #HACK to work around a bug writing to the DC multiple times on OSX
                if "wxOSX" in wx.PlatformInfo:
                    dc = wx.ClientDC(self)
                    self.PrepareDC(dc)
                    # we need to clear the dc BEFORE calling PrepareDC
                    dc.SetUserScale(self.drawScale, self.drawScale)
                
                if e[0][0] >= topLeft[0] and e[0][0] <= bottomRight[0] and e[0][1] <= topLeft[1] and e[0][1] >= bottomRight[1]:
                    if e[1] == SIMFW_DRAWQUEUE_DRAWTILE:
                        loc = self.assembly.GetLoc(e[0])
                        self.DrawAtCoords(dc, e[0][0], e[0][1], loc)
                    elif e[1] == SIMFW_DRAWQUEUE_DRAWFRONTIER:
                        loc = self.assembly.GetLoc(e[0], includeFrontier=True)
                        self.DrawAtCoords(dc, e[0][0], e[0][1], loc)
                    elif (e[1] == SIMFW_DRAWQUEUE_REMOVETILE) or (e[1] == SIMFW_DRAWQUEUE_REMOVEFRONTIER):
                        #HACK to work around a bug writing to the DC multiple times on OSX
                        if not "wxOSX" in wx.PlatformInfo:
                            self.ClearCoord(dc, e[0][0], e[0][1])
                    
            self.DrawQueue = []
        self._lock.release()
        self.needsRedraw = False


        #dc.BeginDrawing()
        #  Iterate the grid and draw tile types
        #for i in range(self.width):
        #    for j in range(self.height):
        #        if self.assembly.grid[i][j] is not None:
        #            self.DrawAtCoords(dc, i, j, self.assembly.grid[i][j])

        #for e in self.RemoveQueue:
        #    dc.RemoveId(e)
            #pen = self.ColoredPen("BLACK")
            #dc.SetPen(pen)
            #dc.SetBrush(self.ColoredBrush("WHITE"))
            #dc.DrawRectangle(0,0,300,300)                         
        #self.RemoveQueue = []
        #dc.EndDrawing()
        
    def DoRedrawAssembly(self):
        self.Refresh()
        
    def InitializeDrawing(self):
        self.DrawAssemblyUpdate("all")
        
#===============================================================================
#     def DoInitializeDrawing(self, dc):
#         #dc.BeginDrawing()
#         
#         bg = wx.Brush("WHITE")#(self.GetBackgroundColour())
#         dc.SetBackground(bg)
#         dc.Clear()
# 
#         self.numVert = self.width / self.sideLen
#         self.numHoriz = self.height / self.sideLen
# 
#         if self.drawGrid == True:
#             dc.SetPen(self.gridPen)
#             for i in range(0, self.numVert+1):
#                 x = i * self.sideLen
#                 dc.DrawLine(x, self.height, x, self.height-self.numVert*self.sideLen)
#             for i in range(0, self.numHoriz+1):
#                 y = i * self.sideLen
#                 dc.DrawLine(0, self.height-y, self.numHoriz*self.sideLen, self.height-y)
#          
#         #dc.EndDrawing()
#         #self.Refresh()
#===============================================================================
        
    def RedrawAssembly(self, checkAll=False):
        if checkAll == True:
            self.DrawAssemblyUpdate("allAndCheck")
        else:
            self.DrawAssemblyUpdate("all")

    def DoScale(self, bIncrease, bScrollToCenter=False, scrollToCoords=None):
        if bScrollToCenter == True:
            coords = self.GetCenterCoords()
        else:
            coords = scrollToCoords
        if bIncrease == True:
            self.drawScale = self.drawScale * 2.0
        else:
            self.drawScale = self.drawScale / 2.0
        print("Scale set to " + str(self.drawScale))
        self.CanvasSizeUpdate(coords)

    def GetPixelsForCoords(self, xCoord, yCoord):
        xCoord = xCoord - self.assembly.minX
        yCoord = yCoord - self.assembly.minY
        xPixel = (xCoord * self.sideLen)# * self.drawScale
        yPixel = (self.height - ((yCoord+1) * self.sideLen))# * self.drawScale
        return (xPixel,yPixel)
  
    def CoordsFromPixelPos(self, x, y, bClosest=True):
        # bClosest isn't used in this default implementation but may be useful in grids where some pixels aren't on valid coordinates
        x1 = int(x / (self.sideLen*self.drawScale))
        y1 = int(((self.height*self.drawScale)-y) / (self.sideLen*self.drawScale))

        x1 = x1 + self.assembly.minX
        y1 = y1 + self.assembly.minY

        return (x1,y1)
    
    def NameMatchesPatterns(self, name):
        if len(self.displayPatterns) == 0:
            return True
        else:
            for pattern in self.displayPatterns:
                if pattern in name:
                    return True
            return False
        
    def DrawFrontierLocation(self, dc, coords, loctype):
        if loctype == TileAssembly.EMPTY_LOC:
            dc.SetBrush(self.whiteBrush)  
        elif loctype == TileAssembly.FRONTIER_LOC_OFF:
            dc.SetBrush(self.redBrush)
        else:
            dc.SetBrush(self.greenBrush)
        (xPixel,yPixel) = self.GetPixelsForCoords(coords[0], coords[1])
        #print("drawing frontier location with center at " + str((xPixel+self.sideLen/2, yPixel+self.sideLen/2)))
        dc.DrawCircle(int(xPixel+self.sideLen/2), int(yPixel+self.sideLen/2), int((self.sideLen/2)-5))

    def ClearCoord(self, dc, xCoord, yCoord):
        (xPixel,yPixel) = self.GetPixelsForCoords(xCoord, yCoord)
        dc.SetBrush(self.backgroundBrush)
        if self.drawGrid == True:
            dc.SetPen(self.gridPen)
        else:
            dc.SetPen(self.backgroundPen)
        dc.DrawRectangle(xPixel, yPixel, self.sideLen, self.sideLen)
        
    def DrawAtCoords(self, dc, xCoord, yCoord, loctype):
        if loctype is None:
            return
        if loctype == TileAssembly.FRONTIER_LOC or loctype == TileAssembly.FRONTIER_LOC_OFF:
            self.DrawFrontierLocation(dc, (xCoord, yCoord), loctype)
        else:
            (xPixel,yPixel) = self.GetPixelsForCoords(xCoord, yCoord)
            # Adjust coords for correct offset of tile images
            xPixel = xPixel - self.imgOffsetHoriz
            yPixel = yPixel - self.imgOffsetVert
            self.DrawFrontierLocation(dc, (xCoord, yCoord), TileAssembly.EMPTY_LOC)
            loctype.type.DrawAtLocation(xPixel, yPixel, dc)

